@@ -1,11 +1,20 @@
-from .lut_tools import read_cal_file, read_cube_file, unity_lut_1d, unity_lut_3d
+from .lut_tools import (
+    create_dolby_vision_config,
+    read_cal_file,
+    read_cube_file,
+    unity_lut_1d,
+    unity_lut_3d,
+    write_dolby_vision_config,
+)
 from .webos_client import PyLGTVCmdException, PyLGTVPairException, WebOsClient
 
 __all__ = [
+    "create_dolby_vision_config",
     "read_cal_file",
     "read_cube_file",
     "unity_lut_1d",
     "unity_lut_3d",
+    "write_dolby_vision_config",
     "PyLGTVCmdException",
     "PyLGTVPairException",
     "WebOsClient",
