@@ -4,12 +4,29 @@ import asyncio
 from aiopylgtv import WebOsClient
 
 
-async def runloop(client, command, parameters):
+async def runloop(args):
+    client = await WebOsClient.create(args.host, timeout_connect=2)
     await client.connect()
-    print(await getattr(client, command)(*parameters))
+    print(await getattr(client, args.command)(*args.parameters))
     await client.disconnect()
 
 
+def convert_arg(arg):
+    try:
+        return int(arg)
+    except ValueError:
+        pass
+    try:
+        return float(arg)
+    except ValueError:
+        pass
+    if arg.lower() == "true":
+        return True
+    elif arg.lower() == "false":
+        return False
+    return arg
+
+
 def aiopylgtvcommand():
     parser = argparse.ArgumentParser(description="Send command to LG WebOs TV.")
     parser.add_argument(
@@ -22,15 +39,11 @@ def aiopylgtvcommand():
     )
     parser.add_argument(
         "parameters",
-        type=str,
+        type=convert_arg,
         nargs="*",
         help="additional parameters to be passed to WebOsClient function call",
     )
 
     args = parser.parse_args()
 
-    client = WebOsClient(args.host, timeout_connect=2)
-
-    asyncio.get_event_loop().run_until_complete(
-        runloop(client, args.command, args.parameters)
-    )
+    asyncio.run(runloop(args))
