@@ -5,6 +5,7 @@ import functools
 import json
 import logging
 import os
+from datetime import timedelta
 
 import numpy as np
 import websockets
@@ -29,6 +30,8 @@ logger = logging.getLogger(__name__)
 KEY_FILE_NAME = ".aiopylgtv.sqlite"
 USER_HOME = "HOME"
 
+SOUND_OUTPUTS_TO_DELAY_CONSECUTIVE_VOLUME_STEPS = {"external_arc"}
+
 
 class PyLGTVPairException(Exception):
     def __init__(self, message):
@@ -59,6 +62,7 @@ class WebOsClient:
         ping_interval=1,
         ping_timeout=20,
         client_key=None,
+        volume_step_delay_ms=None,
     ):
         """Initialize the client."""
         self.ip = ip
@@ -88,8 +92,15 @@ class WebOsClient:
         self._system_info = None
         self._software_info = None
         self._sound_output = None
+        self._picture_settings = None
         self.state_update_callbacks = []
         self.doStateUpdate = False
+        self._volume_step_lock = asyncio.Lock()
+        self._volume_step_delay = (
+            timedelta(milliseconds=volume_step_delay_ms)
+            if volume_step_delay_ms is not None
+            else None
+        )
 
     @classmethod
     async def create(cls, *args, **kwargs):
@@ -261,6 +272,7 @@ class WebOsClient:
                 self.subscribe_apps(self.set_apps_state),
                 self.subscribe_inputs(self.set_inputs_state),
                 self.subscribe_sound_output(self.set_sound_output_state),
+                self.subscribe_picture_settings(self.set_picture_settings_state),
             }
             subscribe_tasks = set()
             for coro in subscribe_coros:
@@ -318,6 +330,7 @@ class WebOsClient:
             self._system_info = None
             self._software_info = None
             self._sound_output = None
+            self._picture_settings = None
 
             for callback in self.state_update_callbacks:
                 closeout.add(callback())
@@ -449,6 +462,10 @@ class WebOsClient:
     def sound_output(self):
         return self._sound_output
 
+    @property
+    def picture_settings(self):
+        return self._picture_settings
+
     @property
     def is_on(self):
         state = self._power_state.get("state")
@@ -631,6 +648,12 @@ class WebOsClient:
         if self.state_update_callbacks and self.doStateUpdate:
             await self.do_state_update_callbacks()
 
+    async def set_picture_settings_state(self, picture_settings):
+        self._picture_settings = picture_settings
+
+        if self.state_update_callbacks and self.doStateUpdate:
+            await self.do_state_update_callbacks()
+
     # low level request handling
 
     async def command(self, request_type, uri, payload=None, uid=None):
@@ -685,7 +708,7 @@ class WebOsClient:
             if error == "404 no such service or method":
                 raise PyLGTVServiceNotFoundError(error)
             else:
-                raise PyLGTVCmdError(error)
+                raise PyLGTVCmdError(response)
         elif returnValue is None:
             raise PyLGTVCmdException(f"Invalid request response {response}")
         elif not returnValue:
@@ -914,11 +937,24 @@ class WebOsClient:
 
     async def volume_up(self):
         """Volume up."""
-        return await self.request(ep.VOLUME_UP)
+        return await self._volume_step(ep.VOLUME_UP)
 
     async def volume_down(self):
         """Volume down."""
-        return await self.request(ep.VOLUME_DOWN)
+        return await self._volume_step(ep.VOLUME_DOWN)
+
+    async def _volume_step(self, endpoint):
+        """Volume step and conditionally sleep afterwards if a consecutive volume step shouldn't be possible to perform immediately after."""
+        if (
+            self.sound_output in SOUND_OUTPUTS_TO_DELAY_CONSECUTIVE_VOLUME_STEPS
+            and self._volume_step_delay is not None
+        ):
+            async with self._volume_step_lock:
+                response = await self.request(endpoint)
+                await asyncio.sleep(self._volume_step_delay.total_seconds())
+                return response
+        else:
+            return await self.request(endpoint)
 
     # TV Channel
     async def channel_up(self):
@@ -1123,6 +1159,282 @@ class WebOsClient:
 
         await self.button(f"""{num}""")
 
+    async def luna_request(self, uri, params):
+        """luna api call."""
+        # n.b. this is a hack which abuses the alert API
+        # to call the internal luna API which is otherwise
+        # not exposed through the websocket interface
+        # An important limitation is that any returned
+        # data is not accessible
+
+        # set desired action for click, fail and close
+        # for redundancy/robustness
+
+        lunauri = f"luna://{uri}"
+
+        buttons = [{"label": "", "onClick": lunauri, "params": params}]
+        payload = {
+            "message": " ",
+            "buttons": buttons,
+            "onclose": {"uri": lunauri, "params": params},
+            "onfail": {"uri": lunauri, "params": params},
+        }
+
+        ret = await self.request(ep.CREATE_ALERT, payload)
+        alertId = ret.get("alertId")
+        if alertId is None:
+            raise PyLGTVCmdException("Invalid alertId")
+
+        return await self.request(ep.CLOSE_ALERT, payload={"alertId": alertId})
+
+    async def set_current_picture_mode(self, pic_mode):
+        """Set picture mode for current input, dynamic range and 3d mode.
+
+        Known picture modes are: cinema, eco, expert1, expert2, game,
+        normal, photo, sports, technicolor, vivid, hdrEffect,  hdrCinema,
+        hdrCinemaBright, hdrExternal, hdrGame, hdrStandard, hdrTechnicolor,
+        hdrVivid, dolbyHdrCinema, dolbyHdrCinemaBright, dolbyHdrDarkAmazon,
+        dolbyHdrGame, dolbyHdrStandard, dolbyHdrVivid, dolbyStandard
+
+        Likely not all modes are valid for all tv models.
+        """
+
+        uri = "com.webos.settingsservice/setSystemSettings"
+
+        params = {"category": "picture", "settings": {"pictureMode": pic_mode}}
+
+        return await self.luna_request(uri, params)
+
+    async def set_picture_mode(
+        self, pic_mode, tv_input, dynamic_range="sdr", stereoscopic="2d"
+    ):
+        """Set picture mode for specific input, dynamic range and 3d mode.
+
+        Known picture modes are: cinema, eco, expert1, expert2, game,
+        normal, photo, sports, technicolor, vivid, hdrEffect,  hdrCinema,
+        hdrCinemaBright, hdrExternal, hdrGame, hdrStandard, hdrTechnicolor,
+        hdrVivid, dolbyHdrCinema, dolbyHdrCinemaBright, dolbyHdrDarkAmazon,
+        dolbyHdrGame, dolbyHdrStandard, dolbyHdrVivid, dolbyStandard
+
+        Known inputs are: atv, av1, av2, camera, comp1, comp2, comp3,
+        default, dtv, gallery, hdmi1, hdmi2, hdmi3, hdmi4,
+        hdmi1_pc, hdmi2_pc, hdmi3_pc, hdmi4_pc, ip, movie,
+        photo, pictest, rgb, scart, smhl
+
+        Known dynamic range modes are: sdr, hdr, technicolorHdr, dolbyHdr
+
+        Known stereoscopic modes are: 2d, 3d
+
+        Likely not all inputs and modes are valid for all tv models.
+        """
+
+        uri = "com.webos.settingsservice/setSystemSettings"
+
+        params = {
+            "category": f"picture${tv_input}.x.{stereoscopic}.{dynamic_range}",
+            "settings": {"pictureMode": pic_mode},
+        }
+
+        return await self.luna_request(uri, params)
+
+    async def set_current_picture_settings(self, settings):
+        """Set picture settings for current picture mode, input, dynamic range and 3d mode.
+
+        A possible list of settings and example values are below (not all settings are applicable
+        for all modes and/or tv models):
+
+        "adjustingLuminance": [
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0
+        ],
+        "backlight": "80",
+        "blackLevel": {
+            "ntsc": "auto",
+            "ntsc443": "auto",
+            "pal": "auto",
+            "pal60": "auto",
+            "palm": "auto",
+            "paln": "auto",
+            "secam": "auto",
+            "unknown": "auto"
+        },
+        "brightness": "50",
+        "color": "50",
+        "colorFilter": "off",
+        "colorGamut": "auto",
+        "colorManagementColorSystem": "red",
+        "colorManagementHueBlue": "0",
+        "colorManagementHueCyan": "0",
+        "colorManagementHueGreen": "0",
+        "colorManagementHueMagenta": "0",
+        "colorManagementHueRed": "0",
+        "colorManagementHueYellow": "0",
+        "colorManagementLuminanceBlue": "0",
+        "colorManagementLuminanceCyan": "0",
+        "colorManagementLuminanceGreen": "0",
+        "colorManagementLuminanceMagenta": "0",
+        "colorManagementLuminanceRed": "0",
+        "colorManagementLuminanceYellow": "0",
+        "colorManagementSaturationBlue": "0",
+        "colorManagementSaturationCyan": "0",
+        "colorManagementSaturationGreen": "0",
+        "colorManagementSaturationMagenta": "0",
+        "colorManagementSaturationRed": "0",
+        "colorManagementSaturationYellow": "0",
+        "colorTemperature": "0",
+        "contrast": "80",
+        "dynamicColor": "off",
+        "dynamicContrast": "off",
+        "edgeEnhancer": "on",
+        "expertPattern": "off",
+        "externalPqlDbType": "none",
+        "gamma": "high2",
+        "grassColor": "0",
+        "hPosition": "0",
+        "hSharpness": "10",
+        "hSize": "0",
+        "hdrDynamicToneMapping": "on",
+        "hdrLevel": "medium",
+        "localDimming": "medium",
+        "motionEyeCare": "off",
+        "motionPro": "off",
+        "mpegNoiseReduction": "off",
+        "noiseReduction": "off",
+        "realCinema": "on",
+        "sharpness": "10",
+        "skinColor": "0",
+        "skyColor": "0",
+        "superResolution": "off",
+        "tint": "0",
+        "truMotionBlur": "10",
+        "truMotionJudder": "0",
+        "truMotionMode": "user",
+        "vPosition": "0",
+        "vSharpness": "10",
+        "vSize": "0",
+        "whiteBalanceApplyAllInputs": "off",
+        "whiteBalanceBlue": [
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0
+        ],
+        "whiteBalanceBlueGain": "0",
+        "whiteBalanceBlueOffset": "0",
+        "whiteBalanceCodeValue": "19",
+        "whiteBalanceColorTemperature": "warm2",
+        "whiteBalanceGreen": [
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0
+        ],
+        "whiteBalanceGreenGain": "0",
+        "whiteBalanceGreenOffset": "0",
+        "whiteBalanceIre": "100",
+        "whiteBalanceLuminance": "130",
+        "whiteBalanceMethod": "2",
+        "whiteBalancePattern": "outer",
+        "whiteBalancePoint": "high",
+        "whiteBalanceRed": [
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0
+        ],
+        "whiteBalanceRedGain": "0",
+        "whiteBalanceRedOffset": "0",
+        "xvycc": "auto"
+
+
+        """
+
+        uri = "com.webos.settingsservice/setSystemSettings"
+
+        params = {"category": "picture", "settings": settings}
+
+        return await self.luna_request(uri, params)
+
+    async def set_picture_settings(
+        self, settings, pic_mode, tv_input, stereoscopic="2d"
+    ):
+        """Set picture settings for specific picture mode, input, and 3d mode."""
+
+        uri = "com.webos.settingsservice/setSystemSettings"
+
+        params = {
+            "category": f"picture${tv_input}.{pic_mode}.{stereoscopic}.x",
+            "settings": settings,
+        }
+
+        return await self.luna_request(uri, params)
+
     def validateCalibrationData(self, data, shape, dtype):
         if not isinstance(data, np.ndarray):
             raise TypeError(f"data must be of type ndarray but is instead {type(data)}")
@@ -1319,6 +1631,15 @@ class WebOsClient:
         ret = await self.request(ep.GET_SYSTEM_SETTINGS, payload=payload)
         return ret["settings"]
 
+    async def subscribe_picture_settings(
+        self, callback, keys=["contrast", "backlight", "brightness", "color"]
+    ):
+        async def settings(payload):
+            await callback(payload.get("settings"))
+
+        payload = {"category": "picture", "keys": keys}
+        return await self.subscribe(settings, ep.GET_SYSTEM_SETTINGS, payload=payload)
+
     async def upload_1d_lut_from_file(self, picMode, filename):
         ext = filename.split(".")[-1].lower()
         if ext == "cal":
